/* vim: set sw=8 ts=8 si et: */
/* $Revision: 1.2 $, last changed: $Date: 1998/06/03 19:17:46 $ File: $RCSfile: bin_asc.c,v $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "common.h"

/*  
 * This converts a binary file into printable ASCII characters, in a
 * radix-64 form mostly compatible with the MIME format.
 * This makes it easier to send encrypted files over a 7-bit channel.
 * 
 * How it works:
 * Take binary data of 3 bytes = 24 bits and devide it into 4 * 6 bits.
 * One 6 bit value is used to generate one ASCII character. Thus
 * 3 binary bytes make 4 ASCII character.
 *
 * This is called radix-64 as 2^6 = 64.
 */
/* Index this array by a 6 bit value to get the character corresponding
 * to that value.
 */
static
unsigned char bintoasc[] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/* Index this array by a 7 bit value to get the 6-bit binary field
 * corresponding to that value.  Any illegal characters return
 * high bit set (0200 = dec 128 = 10000000).
 */
static
unsigned char asctobin[] =
{
        0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0200, 0200, 0076, 0200, 0200, 0200, 0077,
        0064, 0065, 0066, 0067, 0070, 0071, 0072, 0073,
        0074, 0075, 0200, 0200, 0200, 0200, 0200, 0200,
        0200, 0000, 0001, 0002, 0003, 0004, 0005, 0006,
        0007, 0010, 0011, 0012, 0013, 0014, 0015, 0016,
        0017, 0020, 0021, 0022, 0023, 0024, 0025, 0026,
        0027, 0030, 0031, 0200, 0200, 0200, 0200, 0200,
        0200, 0032, 0033, 0034, 0035, 0036, 0037, 0040,
        0041, 0042, 0043, 0044, 0045, 0046, 0047, 0050,
        0051, 0052, 0053, 0054, 0055, 0056, 0057, 0060,
        0061, 0062, 0063, 0200, 0200, 0200, 0200, 0200
};

/* ENC is the basic 1 character encoding function to make a char printing */
#define ENC(c) ((int)bintoasc[((c) & 077)])
#define PAD '='
/*
 * charcters per line (can only be increased in steps of 4)
 * 60 leads to a linelength of 64
 */
#define B64LINELEN 60

/*
 * output one group of up to 3 bytes, pointed at by p, on string buffer 
 * of length 4.
 * If fewer than 3 are present, the 1 or two extras must be zeros.
 * count=1..3
 */
static void do64encode(unsigned char *p, char out[5], int count)
{
        int c1, c2, c3, c4;

        c1 = *p >> 2;
        c2 = ((*p << 4) & 060) | ((p[1] >> 4) & 017);
        c3 = ((p[1] << 2) & 074) | ((p[2] >> 6) & 03);
        c4 = p[2] & 077;
        out[0] = ENC(c1);
        out[1] = ENC(c2);
        if (count == 1) {
                out[2] = PAD;
                out[3] = PAD;
        } else {
                out[2] = ENC(c3);
                if (count == 2)
                        out[3] = PAD;
                else
                        out[3] = ENC(c4);
        }
        out[4] = 0;             /*string termination */
}                               /* outdec */
/* 
 * The public 64 base encoding function. Takes a buffers of type twoword
 * and prints the encoded data to a file. This function maintains a static
 * internal buffer. You must call close64base to flush this buffer before
 * closing the file.
 */
static UBYTE_08bits encbuf[3];
static int encbufpos = 0;
static int charperline = 0;
static int decbuflen = 0;
static int lcount = 0;
static UBYTE_08bits *decbufstartprt;
/*one whole line in binary format must fit into decbuf,
 *B64LINELEN is a bit too much but that's ok.
 */
static UBYTE_08bits decbuf[B64LINELEN + 12];

void open64base_w(FILE * f)
{
        encbufpos = 0;
        charperline = 0;
        if (fputs("...enc.data.start...\n", f) <= 0) {
                fprintf(stderr, "ERROR while writing output 64base output\n");
                exit(1);
        }
}

void write64base(unsigned char bindat[8], FILE * f)
{

        char outbuf[5];

        int i;
        for (i = 0; i < 8; i++) {
                encbuf[encbufpos] = bindat[i];
                encbufpos++;
                if (encbufpos == 3) {
                        do64encode(encbuf, outbuf, encbufpos);
                        if (fputs(outbuf, f) <= 0) {
                                fprintf(stderr, "ERROR while writing output 64base output\n");
                                exit(1);
                        }
                        charperline += 4;
                        encbufpos = 0;
                        if (charperline > B64LINELEN) {
                                fputc('\n', f);
                                charperline = 0;
                        }
                }
        }
}

void close64base_w(FILE * f)
{
        char outbuf[5];
        if (encbufpos) {
                do64encode(encbuf, outbuf, encbufpos);
                if (fputs(outbuf, f) <= 0) {
                        fprintf(stderr, "ERROR while writing output 64base output\n");
                        exit(1);
                }
                encbufpos = 0;
        }
        if (fputs("\n...enc.data.stop...\n", f) <= 0) {
                fprintf(stderr, "ERROR while writing output 64base output\n");
                exit(1);
        }
}


/*
 * Reads a line from file f, up to the size of the buffer.  The line in the
 * buffer will NOT include line termination, although any of (CR, LF, CRLF)
 * is accepted on input.  The return value is 0 on EOF,
 * 2 on out of buffer space and 1 if the line was terminated normally.
 * Tab and spaces are removed from the line.
 *
 * Passing in a buffer less than 2 characters long is not a terribly bright
 * idea.
 */
static int mygetline(char *buf, int buflen, FILE * f)
{
        int state;
        char *p;
        int c;

        state = 0;
        p = buf;

        lcount++; 
        while (1) {
                c = getc(f);
                if (c == '\t' || c == ' ') {
                        continue;
                }
                if (c == '\n') {
                        *p = 0;
                        return 1;       /* Line terminated with \n or \r\n */
                }
                if (state) {
                        ungetc(c, f);
                        *p = 0;
                        return 1;       /* Line terminated with \r only */
                }
                if (c == EOF) {
                        *p = 0;
                        return (0);
                }
                if (c == '\r')
                        state = 1;
                else if (--buflen > 0) {
                        *p++ = c;
                } else {
                        ungetc(c, f);
                        *p = 0;
                        lcount--;
                        return 2;       /* Out of buffer space */
                }
        }
}                               /* mygetline */

/*
 * ASCII base 64 decode routines, The inbuf length must be
 * dividable by 4 without rest. This is the case if it was written
 * with write64base.
 * Return 1 on error, 0 on ok.
 */
static int do64decode(unsigned char *inbuf, unsigned char *outbuf, int *outlength)
{
        unsigned char *bp;
        int length;
        unsigned int c1, c2, c3, c4;
        register int j;

        length = 0;
        bp = inbuf;

        /* FOUR input characters go into each THREE output charcters */

        while (*bp != '\0') {
                if (*bp & 0x80 || (c1 = asctobin[*bp]) & 0x80) {
                        return (1);
                }
                ++bp;
                if (*bp & 0x80 || (c2 = asctobin[*bp]) & 0x80) {
                        return (1);
                }
                if (*++bp == PAD) {
                        c3 = c4 = 0;
                        length += 1;
                        if (strcmp((char *) bp, "==") == 0) {
                                bp += 1;        /*Mime mailers generate this: =3D */
                        } else if (strcmp((char *) bp, "=3D=3D") == 0) {
                                bp += 5;
                        } else {
                                return 1;
                        }
                } else if (*bp & 0x80 || (c3 = asctobin[*bp]) & 0x80) {
                        return 1;
                } else {
                        /*now decode c4:*/
                        if (*++bp == PAD) {
                                c4 = 0;
                                length += 2;
                                if (strcmp((char *) bp, "=") == 0);     /* All is well */
                                else if (strcmp((char *) bp, "=3D") == 0)
                                        bp += 2;
                                else
                                        return 1;
                        } else if (*bp & 0x80 || (c4 = asctobin[*bp]) & 0x80) {
                                return 1;
                        } else {
                                length += 3;
                        }
                }
                ++bp;
                j = (c1 << 2) | (c2 >> 4);
                *outbuf++ = j;
                j = (c2 << 4) | (c3 >> 2);
                *outbuf++ = j;
                j = (c3 << 6) | c4;
                *outbuf++ = j;
        }
        *outlength = length;
        return 0;               /* normal return */

}

/* return 0 if this is an encoded file otherwise return 1 */
int ck64base_r(char *filename)
{
        char tmp[80];
        int notfound = 1;
        FILE *instream;
        instream=fopen(filename,"r");
        if (instream==NULL){
                fprintf(stderr,"ERROR: Could not open file %s for reading\n",filename);
                exit(1);
        }
        lcount = 0;
        while (mygetline(tmp, 80, instream)) {
                if (strcmp(tmp, "...enc.data.start...") == 0) {
                        notfound = 0;
                        break;
                }
                if (lcount > 10) {
                        break;
                }
        }
        fclose(instream);
        return(notfound);
}

void open64base_r(FILE * f)
{
        char tmp[80];
        int notfound = 1;

        decbuflen = 0;
        lcount = 0;
        decbufstartprt = decbuf;
        while (mygetline(tmp, 80, f)) {
                if (strcmp(tmp, "...enc.data.start...") == 0) {
                        notfound = 0;
                        break;
                }
        }
        if (notfound) {
                fprintf(stderr, "ERROR:...enc.data.start... not found\n");
                exit(1);
        }
}
/*
 *64 base decodes file and make binary data available in bindat
 */
int read64base(unsigned char bindat[8], FILE * f)
{
        char tmp[B64LINELEN + 8];
        int i;
        /*decbuflen is end pos of decbuf */
        if (decbuflen < 8) {
                /*copy the remmains to the beginning of decbuf */
                for (i = 0; i < decbuflen; i++) {
                        *(decbuf + i) = *decbufstartprt;
                        decbufstartprt++;
                }
                decbufstartprt = decbuf;
                while ((i = mygetline(tmp, B64LINELEN + 5, f))) {
                        if (*tmp == 0)
                                continue;       /*ignore empty lines */
                        if (strcmp(tmp, "...enc.data.stop...") == 0) {
                                return (0);
                        }
                        break;
                }
                if (i == 0) {
                        /*EOF before stop mark */
                        fprintf(stderr, "ERROR: ...enc.data.stop... not found\n");
                        exit(1);
                }
                if (do64decode((unsigned char*)tmp, decbuf + decbuflen, &i) != 0) {
                        fprintf(stderr, "ERROR: wrong character line %d\n", lcount);
                       fprintf(stderr, "wrong: %s\n",tmp);
                        exit(1);
                }
                decbuflen += i;
        }
        if (decbuflen < 8) {
                return (1);     /*error */
        }
        for (i = 0; i < 8; i++) {
                bindat[i] = *(decbufstartprt);
                decbufstartprt++;
                decbuflen--;
        }
        return (8);             /*normal return, 8 bytes in bindat */

}
