/* vim: set sw=8 ts=8 si et: */
/* $Revision: 1.1 $, last changed: $Date: 1998/05/07 17:54:10 $ File: $RCSfile: blowfish_n.c,v $ */

#include "blowfish.h"
#include "blowfish_tables.h"

#ifdef LLITTLE_ENDIAN            /* Eg: Intel */
#ifdef DOS
#include <alloc.h>
#include <dir.h>
#include <bios.h>
#include <dos.h>
#include <graphics.h>
#include <io.h>
#endif  /* DOS */

#else /* LLITTLE_ENDIAN */
/*BIG_ENDIAN */
#ifdef MAC
#include <Types.h>
#endif /* MAC */
#endif  /* LLITTLE_ENDIAN */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>

#define KEYBYTES         8

/*local to this file: */
static UWORD_32bits P[bf_N + 2];
static UWORD_32bits S[4][256];

#ifdef LLITTLE_ENDIAN            /* Eg: Intel */
static void byteswap(union twoword *two32bit);
#endif  /* LLITTLE_ENDIAN */

static inline UWORD_32bits F(UWORD_32bits x)
{
        UWORD_16bits a;
        UWORD_16bits b;
        UWORD_16bits c;
        UWORD_16bits d;
        UWORD_32bits y;

        d = x & 0x00FF;         /*that is byte-order independent! */
        x >>= 8;
        c = x & 0x00FF;
        x >>= 8;
        b = x & 0x00FF;
        x >>= 8;
        a = x & 0x00FF;
        /*y = ((S[0][a] + S[1][b]) ^ S[2][c]) + S[3][d]; */
        y = S[0][a] + S[1][b];
        y = y ^ S[2][c];
        y = y + S[3][d];

        return y;
}

void do_encipher(UBYTE_08bits *blockof64bit)
{
        union twoword *X;

        UWORD_32bits temp;
        int i;

        X=(union twoword *)blockof64bit;
        /*now we work on two 32 bit numbers, byte swap them*/
#ifdef LLITTLE_ENDIAN 
byteswap(X);
#endif  /* LLITTLE_ENDIAN */

        for (i = 0; i < bf_N; ++i) {
                X->w[0] = X->w[0] ^ P[i];
                X->w[1] = F(X->w[0]) ^ X->w[1];

                temp = X->w[0];
                X->w[0] = X->w[1];
                X->w[1] = temp;
        }

        temp = X->w[0];
        X->w[0] = X->w[1];
        X->w[1] = temp;

        X->w[1] = X->w[1] ^ P[bf_N];
        X->w[0] = X->w[0] ^ P[bf_N + 1];

#ifdef LLITTLE_ENDIAN 
byteswap(X);
#endif  /* LLITTLE_ENDIAN */

}

void do_decipher(UBYTE_08bits *blockof64bit)
{
        union twoword *X;
        UWORD_32bits temp;
        int i;

        X=(union twoword *)blockof64bit;

        /*now we work on 32 bit numbers, byte swap them*/
#ifdef LLITTLE_ENDIAN 
byteswap(X);
#endif  /* LLITTLE_ENDIAN */

        for (i = bf_N + 1; i > 1; --i) {
                X->w[0] = X->w[0] ^ P[i];
                X->w[1] = F(X->w[0]) ^ X->w[1];

                /* Exchange X->w[0] and X->w[1] */
                temp = X->w[0];
                X->w[0] = X->w[1];
                X->w[1] = temp;
        }

        /* Exchange X->w[0] and X->w[1] */
        temp = X->w[0];
        X->w[0] = X->w[1];
        X->w[1] = temp;

        X->w[1] = X->w[1] ^ P[1];
        X->w[0] = X->w[0] ^ P[0];

#ifdef LLITTLE_ENDIAN 
byteswap(X);
#endif  /* LLITTLE_ENDIAN */


}

int cipher_cleanup()
{
        return(0);
}

int Initializekey(unsigned char *key, int keybytes)
{
        int i;
        int j;
        int k;
        UWORD_32bits data;
        union twoword datablk;

        /* First, open the file containing the array initialization data */
        /*copy initial values to working values: */
        for (i = 0; i < bf_N + 2; ++i) {
                P[i] = ini_bf_P[i];
        }
        for (i = 0; i < 4; ++i) {
                for (j = 0; j < 256; ++j) {
                        S[i][j] = ini_bf_S[i][j];
                }
        }
        j = 0;
        for (i = 0; i < bf_N + 2; ++i) {
                data = 0x00000000;
                for (k = 0; k < 4; ++k) {
                        data <<= 8;
                        /*original bug from Bruce Schneier book fixed:*/
                        data |= (UWORD_32bits) key[j] & 0xff;
                        j = j + 1;
                        if (j >= keybytes) {
                                j = 0;
                        }
                }
                P[i] = P[i] ^ data;
        }

        datablk.w[0] = 0x00000000;
        datablk.w[1] = 0x00000000;

        for (i = 0; i < bf_N + 2; i += 2) {
#ifdef LLITTLE_ENDIAN 
byteswap(&datablk);
#endif  /* LLITTLE_ENDIAN */
                do_encipher(datablk.b);
#ifdef LLITTLE_ENDIAN 
byteswap(&datablk);
#endif  /* LLITTLE_ENDIAN */
                P[i] = datablk.w[0];
                P[i + 1] = datablk.w[1];
        }

        for (i = 0; i < 4; ++i) {
                for (j = 0; j < 256; j += 2) {
#ifdef LLITTLE_ENDIAN 
byteswap(&datablk);
#endif  /* LLITTLE_ENDIAN */
                        do_encipher(datablk.b);
#ifdef LLITTLE_ENDIAN 
byteswap(&datablk);
#endif  /* LLITTLE_ENDIAN */
                        S[i][j] = datablk.w[0];
                        S[i][j + 1] = datablk.w[1];
                }
        }
        return (0);
}
/* the byte order of the four bytes ABCD in a 32 bit variable (unsigned long):
   DCBA - little endian - intel 
   ABCD - big endian - motorola, sun */
#ifdef LLITTLE_ENDIAN

/* Byte swap 2  long , used e.g for Intel */
static void byteswap(union twoword *two32bit)
{
        /*now we have 7654 3210 and we need to swap them 
         *to 4567 0123 as we will work with 32 bit values
         */
        char tmp;

        tmp = two32bit->b[3];
        two32bit->b[3] = two32bit->b[0];
        two32bit->b[0] = tmp;

        tmp = two32bit->b[2];
        two32bit->b[2] = two32bit->b[1];
        two32bit->b[1] = tmp;

        tmp = two32bit->b[7];
        two32bit->b[7] = two32bit->b[4];
        two32bit->b[4] = tmp;

        tmp = two32bit->b[6];
        two32bit->b[6] = two32bit->b[5];
        two32bit->b[5] = tmp;
}
#endif  /* LLITTLE_ENDIAN */

#ifdef LLITTLE_ENDIAN_2
/* Byte swap a long , used e.g for Intel */
static void byteswap32bit(UWORD_32bits * data)
{
        *data = ((*data & 0xFF000000) >> 24) |
            ((*data & 0x00FF0000) >> 8) |
            ((*data & 0x0000FF00) << 8) |
            ((*data & 0x000000FF) << 24);
        return data;
}
#endif  /* LLITTLE_ENDIAN */
