/* vim: set sw=8 ts=8 si et: */
/* $Revision: 1.2 $, last changed: $Date: 1998/05/07 17:54:10 $ File: $RCSfile: mygetpass.c,v $ */
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
/* MAXKEYBYTES = max num of char to enter: */
#include "common.h"

#if defined(__msdos) || defined(__MSDOS__)
#include <pc.h>
#include <keys.h>
#define INCLMSDOS
#undef INCLTERMIO
#undef BSDUNIX
#endif


#ifdef BSDUNIX
#include <sgtty.h>
#define TTY     "/dev/tty"      /* Change to "con" for MS-DOS */
#define SIG     void            /* Change to run on sun */
#undef INCLTERMIO
#undef INCLMSDOS
#else
#include <termios.h>
#define INCLTERMIO
#undef BSDUNIX
#undef INCLMSDOS
#endif


/* issue a prompt string, read a key without echo on screen and
 * return it.
 */
char *mygetpass(char *prompt)
{
        static char returned_key_as_string[MAXKEYBYTES + 1];
        int i = 0;
        char c;

#ifdef INCLTERMIO
        sigset_t sig, sigsave;
        struct termios term, termsave;

        sigemptyset(&sig);
        sigaddset(&sig, SIGINT);
        sigaddset(&sig, SIGTSTP);
        sigprocmask(SIG_BLOCK, &sig, &sigsave);
        //
        tcgetattr(fileno(stdin), &termsave);
        term = termsave;
        term.c_lflag &= ~(ECHO | ICANON);
        tcsetattr(fileno(stdin), TCSAFLUSH, &term);
        //
        fprintf(stderr, "%s", prompt);
        fflush(stderr);
        while ((c = getc(stdin)) != EOF && c != '\n') {
                if (c < ' ') {
                        // ignore non printable less than acsii space
                        continue;
                }
                if (c == 0x7f){ // backspace key (sends del)
                        if (i>0){
                                i--;
                                fputc('\b',stderr);
                                fputc(' ',stderr);
                                fputc('\b',stderr);
                        }
                        continue;
                }
                if(i < MAXKEYBYTES){
                        returned_key_as_string[i] = c;
                        i++;
                        fprintf(stderr, "*");
                        fflush(stderr);
                }
        }
        returned_key_as_string[i] = '\0';
        fprintf(stderr, "\n");
        fflush(stderr);
        tcsetattr(fileno(stdin), TCSAFLUSH, &termsave);
        sigprocmask(SIG_SETMASK, &sigsave, NULL);
        if (i<3) {
                fprintf(stderr, "Error: key too short.\n");
                exit(1);
        }
        return (returned_key_as_string);

#endif // INCLTERMIO 
#ifdef INCLMSDOS
/* this code works for gcc under dos: */

        fprintf(stderr, "%s", prompt);
        fflush(stderr);
        while ((c = getkey()) != K_Return) {
                if(i < MAXKEYBYTES){
                        returned_key_as_string[i] = c;
                        i++;
                        fprintf(stderr, "*");
                        fflush(stderr);
                }
        }
        returned_key_as_string[i] = '\0';
        fprintf(stderr, "\r\n");
        if (i<3) {
                fprintf(stderr, "Error: key too short.\r\n");
                exit(1);
        }
        return (returned_key_as_string);
#endif  /* INCLMSDOS */
#ifdef BSDUNIX
        /*this part is not very well tested and may cause problems */
        struct sgttyb ttyb, ttysav;
        register char *cp;
        FILE *tty;
        SIG(*signal())(), (*sig) ();    /* change for sun */

        if ((tty = fdopen(open(TTY, 2), "r")) == NULL)
                tty = stdin;
        else
                setbuf(tty, (char *) NULL);
        sig = signal(SIGINT, SIG_IGN);
        ioctl(fileno(tty), TIOCGETP, &ttyb);
        ioctl(fileno(tty), TIOCGETP, &ttysav);
        ttyb.sg_flags |= RAW;
        ttyb.sg_flags &= ~(ECHO|ICANON);
        ioctl(fileno(tty), TIOCSETP, &ttyb);
        fprintf(stderr, "%s", prompt);
        fflush(stderr);
        cp = returned_key_as_string;
        for (;;) {
                c = getc(tty);
                if (c == '\r' || c == '\n' || c == EOF)
                        break;
                if(i < MAXKEYBYTES){
                        returned_key_as_string[i] = c;
                        i++;
                        fprintf(stderr, "*");
                        fflush(stderr);
                }
        }
        returned_key_as_string[i] = '\0';
        fprintf(stderr,"\n");
        fflush(stderr);
        ioctl(fileno(tty), TIOCSETP, &ttysav);
        signal(SIGINT, sig);
        if (tty != stdin)
                fclose(tty);
        if (i<3) {
                fprintf(stderr,"Error: key too short.\n");
                exit(1);
        }
        return (returned_key_as_string);
#endif  /* BSDUNIX */
}
